@@ -265,9 +265,21 @@ enum ie_type {
 	IE_TYPE_FILS_PUBLIC_KEY                      = 256 + 12,
 	IE_TYPE_FILS_NONCE                           = 256 + 13,
 	IE_TYPE_FUTURE_CHANNEL_GUIDANCE              = 256 + 14,
+	IE_TYPE_SERVICE_HINT                         = 256 + 15,
+	IE_TYPE_SERVICE_HASH                         = 256 + 16,
+	IE_TYPE_CDMG_CAPABILITIES                    = 256 + 17,
+	IE_TYPE_CLUSTER_PROBE                        = 256 + 21,
+	IE_TYPE_CMMG_CAPABILITIES                    = 256 + 27,
+	IE_TYPE_CMMG_OPERATION                       = 256 + 28,
 	IE_TYPE_OWE_DH_PARAM                         = 256 + 32,
 	IE_TYPE_PASSWORD_IDENTIFIER                  = 256 + 33,
+	IE_TYPE_GLK_GCR_PARAMETER_SET                = 256 + 34,
+	IE_TYPE_VENDOR_SPECIFIC_REQUEST              = 256 + 44,
+	IE_TYPE_MAX_CHANNEL_SWITCH_TIME              = 256 + 52,
+	IE_TYPE_ESTIMATED_SERVICE_PARAMETERS_OUT     = 256 + 53,
 	IE_TYPE_OCI                                  = 256 + 54,
+	IE_TYPE_MSCS_DESCRIPTOR                      = 256 + 88,
+	IE_TYPE_SUPPLEMENTAL_CLASS_2_CAPABILITIES    = 256 + 90,
 	IE_TYPE_REJECTED_GROUPS                      = 256 + 92,
 	IE_TYPE_ANTI_CLOGGING_TOKEN_CONTAINER        = 256 + 93,
 };
