@@ -124,9 +124,10 @@ enum mmpdu_reason_code {
 };
 
 /*
- * 802.11-2016, Section 9.4.1.9 Status Code field
+ * 802.11-2020, Section 9.4.1.9 Status Code field
  */
 enum mmpdu_status_code {
+	MMPDU_STATUS_CODE_SUCCESS = 0,
 	MMPDU_STATUS_CODE_UNSPECIFIED = 1,
 	MMPDU_STATUS_CODE_TDLS_REJECTED_ALT_PROV = 2,
 	MMPDU_STATUS_CODE_TDLS_REJECTED = 3,
@@ -230,6 +231,7 @@ enum mmpdu_status_code {
 	MMPDU_STATUS_CODE_ENABLEMENT_DENIED = 105,
 	MMPDU_STATUS_CODE_RESTRICT_AUTH_GDB = 106,
 	MMPDU_STATUS_CODE_AUTHORIZATION_DEENABLED = 107,
+	MMPDU_STATUS_CODE_SAE_HASH_TO_ELEMENT = 126,
 };
 
 /* 802.11, Section 8.2.4.1.1, Figure 8-2 */
@@ -409,3 +411,6 @@ struct mmpdu_deauthentication {
 const struct mmpdu_header *mpdu_validate(const uint8_t *frame, int len);
 const void *mmpdu_body(const struct mmpdu_header *mpdu);
 size_t mmpdu_header_len(const struct mmpdu_header *mmpdu);
+
+int mpdu_sort_ies(enum mpdu_management_subtype type,
+			struct iovec *iov, size_t n_iovs);
