@@ -32,8 +32,11 @@
 #include "src/handshake.h"
 #include "src/crypto.h"
 #include "src/mpdu.h"
-#include "src/sae.h"
 #include "src/auth-proto.h"
+#include "src/sae.h"
+
+/* SHA-512 is the highest supported hashing function as of 802.11-2020 */
+#define SAE_MAX_HASH_LEN 64
 
 #define SAE_RETRANSMIT_TIMEOUT	2
 #define SAE_SYNC_MAX		3
@@ -53,15 +56,15 @@ struct sae_sm {
 	enum sae_state state;
 	const struct l_ecc_curve *curve;
 	unsigned int group;
-	uint8_t group_retry;
-	const unsigned int *ecc_groups;
+	int group_retry;
+	uint16_t *rejected_groups;
 	struct l_ecc_scalar *rand;
 	struct l_ecc_scalar *scalar;
 	struct l_ecc_scalar *p_scalar;
 	struct l_ecc_point *element;
 	struct l_ecc_point *p_element;
 	uint16_t send_confirm;
-	uint8_t kck[32];
+	uint8_t kck[SAE_MAX_HASH_LEN];
 	uint8_t pmk[32];
 	uint8_t pmkid[16];
 	uint8_t *token;
@@ -79,8 +82,99 @@ struct sae_sm {
 	sae_tx_authenticate_func_t tx_auth;
 	sae_tx_associate_func_t tx_assoc;
 	void *user_data;
+	enum crypto_sae sae_type;
 };
 
+static enum mmpdu_status_code sae_status_code(struct sae_sm *sm)
+{
+	switch (sm->sae_type) {
+	case CRYPTO_SAE_LOOPING:
+		return MMPDU_STATUS_CODE_SUCCESS;
+	case CRYPTO_SAE_HASH_TO_ELEMENT:
+		return MMPDU_STATUS_CODE_SAE_HASH_TO_ELEMENT;
+	}
+
+	return MMPDU_STATUS_CODE_UNSPECIFIED;
+}
+
+static void sae_rejected_groups_append(struct sae_sm *sm, uint16_t group)
+{
+	uint16_t i;
+
+	if (!sm->rejected_groups) {
+		sm->rejected_groups = reallocarray(NULL, 2, sizeof(uint16_t));
+		sm->rejected_groups[0] = 1;
+		sm->rejected_groups[1] = group;
+		return;
+	}
+
+	for (i = 1; i <= sm->rejected_groups[0]; i++)
+		if (sm->rejected_groups[i] == group)
+			return;
+
+	sm->rejected_groups = reallocarray(sm->rejected_groups,
+						i + 1, sizeof(uint16_t));
+	sm->rejected_groups[0] += 1;
+	sm->rejected_groups[i] = group;
+}
+
+static void sae_reset_state(struct sae_sm *sm)
+{
+	l_ecc_scalar_free(sm->scalar);
+	sm->scalar = NULL;
+	l_ecc_scalar_free(sm->p_scalar);
+	sm->p_scalar = NULL;
+	l_ecc_scalar_free(sm->rand);
+	sm->rand = NULL;
+	l_ecc_point_free(sm->element);
+	sm->element = NULL;
+	l_ecc_point_free(sm->p_element);
+	sm->p_element = NULL;
+	l_ecc_point_free(sm->pwe);
+	sm->pwe = NULL;
+}
+
+static int sae_choose_next_group(struct sae_sm *sm)
+{
+	const unsigned int *ecc_groups = l_ecc_supported_ike_groups();
+	bool reset = sm->group_retry >= 0;
+
+	do {
+		sm->group_retry++;
+
+		if (ecc_groups[sm->group_retry] == 0)
+			return -ENOENT;
+	} while (sm->sae_type != CRYPTO_SAE_LOOPING &&
+			!sm->handshake->ecc_sae_pts[sm->group_retry]);
+
+	if (reset)
+		sae_reset_state(sm);
+
+	sm->group = ecc_groups[sm->group_retry];
+	sm->curve = l_ecc_curve_from_ike_group(sm->group);
+
+	return 0;
+}
+
+static int sae_valid_group(struct sae_sm *sm, unsigned int group)
+{
+	const unsigned int *ecc_groups = l_ecc_supported_ike_groups();
+	unsigned int i;
+
+	for (i = sm->group_retry; ecc_groups[i]; i++) {
+		if (ecc_groups[i] != group)
+			continue;
+
+		if (sm->sae_type != CRYPTO_SAE_LOOPING &&
+				!sm->handshake->ecc_sae_pts[i])
+			continue;
+
+		return i;
+	}
+
+	return -ENOENT;
+}
+
 static bool sae_pwd_seed(const uint8_t *addr1, const uint8_t *addr2,
 				uint8_t *base, size_t base_len,
 				uint8_t counter, uint8_t *out)
@@ -145,25 +239,31 @@ static struct l_ecc_scalar *sae_pwd_value(const struct l_ecc_curve *curve,
 }
 
 /* IEEE 802.11-2016 - Section 12.4.2 Assumptions on SAE */
-static bool sae_cn(const uint8_t *kck, uint16_t send_confirm,
+static ssize_t sae_cn(struct sae_sm *sm, uint16_t send_confirm,
 			struct l_ecc_scalar *scalar1,
 			struct l_ecc_point *element1,
 			struct l_ecc_scalar *scalar2,
 			struct l_ecc_point *element2,
 			uint8_t *confirm)
 {
+	enum l_checksum_type hash =
+		crypto_sae_hash_from_ecc_prime_len(sm->sae_type,
+				l_ecc_curve_get_scalar_bytes(sm->curve));
+	size_t hash_len = l_checksum_digest_length(hash);
 	uint8_t s1[L_ECC_SCALAR_MAX_BYTES];
 	uint8_t s2[L_ECC_SCALAR_MAX_BYTES];
 	uint8_t e1[L_ECC_POINT_MAX_BYTES];
 	uint8_t e2[L_ECC_POINT_MAX_BYTES];
 	struct l_checksum *hmac;
 	struct iovec iov[5];
-	int ret;
+	ssize_t ret;
 
-	hmac = l_checksum_new_hmac(L_CHECKSUM_SHA256, kck, 32);
+	hmac = l_checksum_new_hmac(hash, sm->kck, hash_len);
 	if (!hmac)
 		return false;
 
+	send_confirm = L_CPU_TO_LE16(send_confirm);
+
 	iov[0].iov_base = &send_confirm;
 	iov[0].iov_len = 2;
 	iov[1].iov_base = (void *) s1;
@@ -176,32 +276,35 @@ static bool sae_cn(const uint8_t *kck, uint16_t send_confirm,
 	iov[4].iov_len = l_ecc_point_get_data(element2, e2, sizeof(e2));
 
 	l_checksum_updatev(hmac, iov, 5);
-
-	ret = l_checksum_get_digest(hmac, confirm, 32);
-
+	ret = l_checksum_get_digest(hmac, confirm, hash_len);
 	l_checksum_free(hmac);
 
-	return (ret == 32);
+	return ret;
 }
 
-static void sae_reject_authentication(struct sae_sm *sm, uint16_t reason)
+static int sae_reject(struct sae_sm *sm, uint16_t transaction, uint16_t status)
 {
 	uint8_t reject[6];
 	uint8_t *ptr = reject;
 
+	if (!sm->handshake->authenticator)
+		return -EPROTO;
+
 	/* transaction */
-	l_put_u16(1, ptr);
-	ptr += 2;
-	/* status success */
-	l_put_u16(reason, ptr);
+	l_put_u16(transaction, ptr);
 	ptr += 2;
 
-	if (reason == MMPDU_STATUS_CODE_UNSUPP_FINITE_CYCLIC_GROUP) {
+	l_put_u16(status, ptr);
+	ptr += 2;
+
+	if (status == MMPDU_STATUS_CODE_UNSUPP_FINITE_CYCLIC_GROUP) {
 		l_put_u16(sm->group, ptr);
 		ptr += 2;
 	}
 
 	sm->tx_auth(reject, ptr - reject, sm->user_data);
+
+	return status;
 }
 
 static struct l_ecc_scalar *sae_new_residue(const struct l_ecc_curve *curve,
@@ -268,8 +371,10 @@ static uint8_t sae_is_quadradic_residue(const struct l_ecc_curve *curve,
  * IEEE 802.11-2016 Section 12.4.4.2.2
  * Generation of the password element with ECC groups
  */
-static bool sae_compute_pwe(struct sae_sm *sm, char *password,
-				const uint8_t *addr1, const uint8_t *addr2)
+static struct l_ecc_point *sae_compute_pwe(const struct l_ecc_curve *curve,
+						const char *password,
+						const uint8_t *addr1,
+						const uint8_t *addr2)
 {
 	uint8_t found = 0;
 	uint8_t is_residue;
@@ -285,10 +390,11 @@ static bool sae_compute_pwe(struct sae_sm *sm, char *password,
 	struct l_ecc_scalar *qr;
 	struct l_ecc_scalar *qnr;
 	uint8_t qnr_bin[L_ECC_SCALAR_MAX_BYTES] = {0};
+	struct l_ecc_point *pwe;
 
 	/* create qr/qnr prior to beginning hunting-and-pecking loop */
-	qr = sae_new_residue(sm->curve, true);
-	qnr = sae_new_residue(sm->curve, false);
+	qr = sae_new_residue(curve, true);
+	qnr = sae_new_residue(curve, false);
 	l_ecc_scalar_get_data(qnr, qnr_bin, sizeof(qnr_bin));
 
 	/*
@@ -326,13 +432,13 @@ static bool sae_compute_pwe(struct sae_sm *sm, char *password,
 		 * execution can continue whatever the result is, without
 		 * changing the outcome.
 		 */
-		pwd_value = sae_pwd_value(sm->curve, pwd_seed, qnr_bin);
+		pwd_value = sae_pwd_value(curve, pwd_seed, qnr_bin);
 
 		/*
 		 * Check if the candidate is a valid x-coordinate on our curve,
 		 * and convert it from scalar to binary.
 		 */
-		is_residue = sae_is_quadradic_residue(sm->curve, pwd_value,
+		is_residue = sae_is_quadradic_residue(curve, pwd_value,
 								qr, qnr);
 		l_ecc_scalar_get_data(pwd_value, x_cand, sizeof(x_cand));
 
@@ -362,37 +468,46 @@ static bool sae_compute_pwe(struct sae_sm *sm, char *password,
 
 	if (!found) {
 		l_error("max PWE iterations reached!");
-		return false;
+		return NULL;
 	}
 
-	sm->pwe = l_ecc_point_from_data(sm->curve, !is_odd + 2, x, sizeof(x));
-	if (!sm->pwe) {
+	pwe = l_ecc_point_from_data(curve, !is_odd + 2, x, sizeof(x));
+	if (!pwe)
 		l_error("computing y failed, was x quadratic residue?");
-		return false;
-	}
 
-	return true;
+	return pwe;
 }
 
-static bool sae_build_commit(struct sae_sm *sm, const uint8_t *addr1,
+static int sae_build_commit(struct sae_sm *sm, const uint8_t *addr1,
 				const uint8_t *addr2, uint8_t *commit,
-				size_t *len, bool retry)
+				size_t len, bool retry)
 {
 	struct l_ecc_scalar *mask;
 	uint8_t *ptr = commit;
 	struct l_ecc_scalar *order;
+	struct ie_tlv_builder builder;
 
 	if (retry)
 		goto old_commit;
 
-	if (!sm->handshake->passphrase) {
-		l_error("no handshake passphrase found");
-		return false;
+	switch (sm->sae_type) {
+	case CRYPTO_SAE_HASH_TO_ELEMENT:
+	{
+		const struct l_ecc_point *pt =
+			sm->handshake->ecc_sae_pts[sm->group_retry];
+
+		sm->pwe = crypto_derive_sae_pwe_from_pt_ecc(addr1, addr2, pt);
+		break;
+	}
+	case CRYPTO_SAE_LOOPING:
+		sm->pwe = sae_compute_pwe(sm->curve, sm->handshake->passphrase,
+						addr1, addr2);
+		break;
 	}
 
-	if (!sae_compute_pwe(sm, sm->handshake->passphrase, addr1, addr2)) {
+	if (!sm->pwe) {
 		l_error("could not compute PWE");
-		return false;
+		return -EIO;
 	}
 
 	sm->scalar = l_ecc_scalar_new(sm->curve, NULL, 0);
@@ -420,17 +535,25 @@ static bool sae_build_commit(struct sae_sm *sm, const uint8_t *addr1,
 	 */
 old_commit:
 
-	/* transaction */
+	/*
+	 * 12.4.7.4 Encoding and decoding of SAE Commit messages
+	 * Refer to Table 9-40 for order and Table 9-41 for presence
+	 * of elements
+	 */
+
+	/* "a Transaction Sequence Number of 1" */
 	l_put_le16(1, ptr);
 	ptr += 2;
-	/* status success */
-	l_put_le16(0, ptr);
+
+	/* "a Status Code of SUCCESS or SAE_HASH_TO_ELEMENT" */
+	l_put_le16(sae_status_code(sm), ptr);
 	ptr += 2;
+
 	/* group */
 	l_put_le16(sm->group, ptr);
 	ptr += 2;
 
-	if (sm->token) {
+	if (sm->sae_type == CRYPTO_SAE_LOOPING && sm->token) {
 		memcpy(ptr, sm->token, sm->token_len);
 		ptr += sm->token_len;
 	}
@@ -438,23 +561,40 @@ old_commit:
 	ptr += l_ecc_scalar_get_data(sm->scalar, ptr, L_ECC_SCALAR_MAX_BYTES);
 	ptr += l_ecc_point_get_data(sm->element, ptr, L_ECC_POINT_MAX_BYTES);
 
-	*len = ptr - commit;
+	ie_tlv_builder_init(&builder, ptr, len - (ptr - commit));
 
-	return true;
+	if (sm->sae_type != CRYPTO_SAE_LOOPING && sm->rejected_groups) {
+		ie_tlv_builder_next(&builder, IE_TYPE_REJECTED_GROUPS);
+		ie_tlv_builder_set_data(&builder, sm->rejected_groups + 1,
+				sm->rejected_groups[0] * sizeof(uint16_t));
+	}
+
+	if (sm->sae_type != CRYPTO_SAE_LOOPING && sm->token) {
+		ie_tlv_builder_next(&builder,
+					IE_TYPE_ANTI_CLOGGING_TOKEN_CONTAINER);
+		ie_tlv_builder_set_data(&builder, sm->token, sm->token_len);
+	}
+
+	ie_tlv_builder_finalize(&builder, &len);
+
+	return ptr - commit + len;
 }
 
-static void sae_send_confirm(struct sae_sm *sm)
+static bool sae_send_confirm(struct sae_sm *sm)
 {
-	uint8_t confirm[32];
-	uint8_t body[38];
+	uint8_t confirm[SAE_MAX_HASH_LEN];
+	uint8_t body[sizeof(confirm) + 6];
 	uint8_t *ptr = body;
+	ssize_t r;
 
 	/*
 	 * confirm = CN(KCK, send-confirm, commit-scalar, COMMIT-ELEMENT,
 	 *			peer-commit-scalar, PEER-COMMIT-ELEMENT)
 	 */
-	sae_cn(sm->kck, sm->sc, sm->scalar, sm->element, sm->p_scalar,
+	r = sae_cn(sm, sm->sc, sm->scalar, sm->element, sm->p_scalar,
 			sm->p_element, confirm);
+	if (r < 0)
+		return false;
 
 	l_put_le16(2, ptr);
 	ptr += 2;
@@ -462,56 +602,39 @@ static void sae_send_confirm(struct sae_sm *sm)
 	ptr += 2;
 	l_put_le16(sm->sc, ptr);
 	ptr += 2;
-	memcpy(ptr, confirm, 32);
-	ptr += 32;
+	memcpy(ptr, confirm, r);
+	ptr += r;
 
-	sm->state = SAE_STATE_CONFIRMED;
-
-	sm->tx_auth(body, 38, sm->user_data);
+	sm->tx_auth(body, ptr - body, sm->user_data);
+	return true;
 }
 
 static int sae_process_commit(struct sae_sm *sm, const uint8_t *from,
 					const uint8_t *frame, size_t len)
 {
 	uint8_t *ptr = (uint8_t *) frame;
-	uint8_t k[L_ECC_SCALAR_MAX_BYTES];
+	unsigned int nbytes = l_ecc_curve_get_scalar_bytes(sm->curve);
+	enum l_checksum_type hash =
+		crypto_sae_hash_from_ecc_prime_len(sm->sae_type, nbytes);
+	size_t hash_len = l_checksum_digest_length(hash);
 	struct l_ecc_point *k_point;
-	uint8_t zero_key[32] = { 0 };
-	uint8_t keyseed[32];
-	uint8_t kck_and_pmk[2][32];
+	uint8_t k[L_ECC_SCALAR_MAX_BYTES];
+	ssize_t klen;
+	const void *salt = NULL;
+	size_t salt_len = 0;
+	uint8_t keyseed[SAE_MAX_HASH_LEN];
+	uint8_t kck_and_pmk[SAE_MAX_HASH_LEN + 32];
 	uint8_t tmp[L_ECC_SCALAR_MAX_BYTES];
 	struct l_ecc_scalar *tmp_scalar;
-	uint16_t group;
-	uint16_t reason = MMPDU_REASON_CODE_UNSPECIFIED;
-	ssize_t klen;
 	struct l_ecc_scalar *order;
-	unsigned int nbytes = l_ecc_curve_get_scalar_bytes(sm->curve);
 
-	if (sm->state != SAE_STATE_COMMITTED) {
-		l_error("bad state %u", sm->state);
-		goto reject;
-	}
-
-	/* Scalar + Point + group */
-	if (len < nbytes + nbytes * 2 + 2) {
-		l_error("bad packet length");
-		goto reject;
-	}
-
-	group = l_get_le16(ptr);
 	ptr += 2;
 
-	if (group != sm->group) {
-		sae_reject_authentication(sm,
-				MMPDU_STATUS_CODE_UNSUPP_FINITE_CYCLIC_GROUP);
-		return 0;
-	}
-
 	sm->p_scalar = l_ecc_scalar_new(sm->curve, ptr, nbytes);
 	if (!sm->p_scalar) {
 		l_error("Server sent invalid P_Scalar during commit");
-		reason = MMPDU_REASON_CODE_UNSPECIFIED;
-		goto reject;
+		return sae_reject(sm, SAE_STATE_COMMITTED,
+				MMPDU_STATUS_CODE_UNSUPP_FINITE_CYCLIC_GROUP);
 	}
 
 	ptr += nbytes;
@@ -520,16 +643,20 @@ static int sae_process_commit(struct sae_sm *sm, const uint8_t *from,
 						ptr, nbytes * 2);
 	if (!sm->p_element) {
 		l_error("Server sent invalid P_Element during commit");
-		reason = MMPDU_REASON_CODE_UNSPECIFIED;
-		goto reject;
+		return sae_reject(sm, SAE_STATE_COMMITTED,
+				MMPDU_STATUS_CODE_UNSUPP_FINITE_CYCLIC_GROUP);
 	}
 
+	/*
+	 * If they match those sent as part of the protocol instance's own
+	 * SAE Commit message, the frame shall be silently discarded (because
+	 * it is evidence of a reflection attack) and the t0 (retransmission)
+	 * timer shall be set.
+	 */
 	if (l_ecc_scalars_are_equal(sm->p_scalar, sm->scalar) ||
 			l_ecc_points_are_equal(sm->p_element, sm->element)) {
-		/* possible reflection attack, silently discard message */
 		l_warn("peer scalar or element matched own, discarding frame");
-
-		return 0;
+		return -ENOMSG;
 	}
 
 	sm->sc++;
@@ -556,18 +683,46 @@ static int sae_process_commit(struct sae_sm *sm, const uint8_t *from,
 	 * i.e., if P = (x, y) then F(P) = x.
 	 */
 	klen = l_ecc_point_get_x(k_point, k, sizeof(k));
-
 	l_ecc_point_free(k_point);
 
 	if (klen < 0)
-		goto reject;
-
-	/* keyseed = H(<0>32, k) */
-	hmac_sha256(zero_key, 32, k, klen, keyseed, 32);
+		return sae_reject(sm, SAE_STATE_COMMITTED,
+				MMPDU_STATUS_CODE_UNSPECIFIED);
 
 	/*
-	 * kck_and_pmk = KDF-Hash-512(keyseed, "SAE KCK and PMK",
-				(commit-scalar + peer-commit-scalar) mod r)
+	 * keyseed = H(salt, k)
+	 *
+	 * 802.11-2020 12.4.5.4:
+	 * Hash to Element case:
+	 * "... a salt consisting of the concatenation of the rejected groups
+	 * from each peer's Rejected Groups element shall be
+	 * passed to the KDF; those of the peer with the highest MAC address go
+	 * first (if only one sent a Rejected Groups element then the salt will
+	 * consist of that list). "
+	 *
+	 * Looping case:
+	 * "...the salt shall consist of a series of octets of the value zero
+	 * whose length equals the length of the digest of the hash function
+	 * used to instantiate H()."
+	 *
+	 * NOTE: We use hkdf_extract here since it is just an hmac invocation
+	 * and it handles the case of the zero key for us.
+	*/
+	if (sm->sae_type != CRYPTO_SAE_LOOPING && sm->rejected_groups) {
+		salt = sm->rejected_groups + 1;
+		salt_len = sm->rejected_groups[0] * sizeof(uint16_t);
+	}
+
+	hkdf_extract(hash, salt, salt_len, 1, keyseed, k, klen);
+
+	/*
+	 * context = (commit-scalar + peer-commit-scalar) mod r
+	 * Length = Q + 256
+	 * kck_and_pmk = KDF-Hash-Length(keyseed, "SAE KCK and PMK", context)
+	 * KCK = L(kck_and_pmk, 0, Q)
+	 * PMK = L(kck_and_pmk, Q, 256)
+	 *
+	 * Q is the length of the digest of the H(), the hash function used
 	 */
 	tmp_scalar = l_ecc_scalar_new(sm->curve, NULL, 0);
 	order = l_ecc_curve_get_order(sm->curve);
@@ -575,11 +730,12 @@ static int sae_process_commit(struct sae_sm *sm, const uint8_t *from,
 	l_ecc_scalar_add(tmp_scalar, sm->p_scalar, sm->scalar, order);
 	l_ecc_scalar_get_data(tmp_scalar, tmp, sizeof(tmp));
 
-	kdf_sha256(keyseed, 32, "SAE KCK and PMK", strlen("SAE KCK and PMK"),
-			tmp, nbytes, kck_and_pmk, 64);
+	crypto_kdf(hash, keyseed, hash_len,
+			"SAE KCK and PMK", strlen("SAE KCK and PMK"),
+			tmp, nbytes, kck_and_pmk, hash_len + 32);
 
-	memcpy(sm->kck, kck_and_pmk[0], 32);
-	memcpy(sm->pmk, kck_and_pmk[1], 32);
+	memcpy(sm->kck, kck_and_pmk, hash_len);
+	memcpy(sm->pmk, kck_and_pmk + hash_len, 32);
 
 	/*
 	 * PMKID = L((commit-scalar + peer-commit-scalar) mod r, 0, 128)
@@ -593,27 +749,27 @@ static int sae_process_commit(struct sae_sm *sm, const uint8_t *from,
 	/* don't set the handshakes pmkid until confirm is verified */
 	memcpy(sm->pmkid, tmp, 16);
 
-	sae_send_confirm(sm);
+	if (!sae_send_confirm(sm))
+		return -EPROTO;
+
+	sm->state = SAE_STATE_CONFIRMED;
 
 	return 0;
-
-reject:
-	sae_reject_authentication(sm, reason);
-	return -EBADMSG;
 }
 
 static bool sae_verify_confirm(struct sae_sm *sm, const uint8_t *frame)
 {
-	uint8_t check[32];
+	uint8_t check[SAE_MAX_HASH_LEN];
 	uint16_t rc = l_get_le16(frame);
+	ssize_t r;
 
-	sae_cn(sm->kck, rc, sm->p_scalar, sm->p_element, sm->scalar,
+	r = sae_cn(sm, rc, sm->p_scalar, sm->p_element, sm->scalar,
 			sm->element, check);
+	if (r < 0)
+		return false;
 
-	if (memcmp(frame + 2, check, 32)) {
-		l_error("confirm did not match");
+	if (memcmp(frame + 2, check, r))
 		return false;
-	}
 
 	sm->rc = rc;
 
@@ -625,19 +781,18 @@ static int sae_process_confirm(struct sae_sm *sm, const uint8_t *from,
 {
 	const uint8_t *ptr = frame;
 
-	if (sm->state != SAE_STATE_CONFIRMED) {
-		l_error("bad state %u", sm->state);
-		goto reject;
+	/*
+	 * If processing is unsuccessful and the SAE Confirm message is not
+	 * verified, protocol instance shall remain in Confirmed state.
+	 *
+	 * NOTE: We diverge from the protocol here and bail out early
+	 */
+	if (!sae_verify_confirm(sm, ptr)) {
+		l_error("SAE: Confirm could not be verified");
+		return sae_reject(sm, SAE_STATE_CONFIRMED,
+					MMPDU_STATUS_CODE_UNSPECIFIED);
 	}
 
-	if (len < 34) {
-		l_error("bad length");
-		goto reject;
-	}
-
-	if (!sae_verify_confirm(sm, ptr))
-		goto reject;
-
 	/* Sc shall be set to the value 2^16 - 1 */
 	sm->sc = 0xffff;
 
@@ -649,25 +804,21 @@ static int sae_process_confirm(struct sae_sm *sm, const uint8_t *from,
 	sm->tx_assoc(sm->user_data);
 
 	return 0;
-
-reject:
-	sae_reject_authentication(sm, MMPDU_REASON_CODE_UNSPECIFIED);
-	return -EBADMSG;
 }
 
 static bool sae_send_commit(struct sae_sm *sm, bool retry)
 {
 	struct handshake_state *hs = sm->handshake;
-	/* regular commit + possible 256 byte token + 6 bytes header */
-	uint8_t commit[L_ECC_SCALAR_MAX_BYTES + L_ECC_POINT_MAX_BYTES + 262];
-	size_t len;
+	/* regular commit + 3x IEs (257 bytes) + 6 bytes header */
+	uint8_t commit[L_ECC_SCALAR_MAX_BYTES + L_ECC_POINT_MAX_BYTES + 777];
+	int r;
 
-	if (!sae_build_commit(sm, hs->spa, hs->aa, commit, &len, retry))
+	r = sae_build_commit(sm, hs->spa, hs->aa,
+					commit, sizeof(commit), retry);
+	if (r < 0)
 		return false;
 
-	sm->state = SAE_STATE_COMMITTED;
-
-	sm->tx_auth(commit, len, sm->user_data);
+	sm->tx_auth(commit, r, sm->user_data);
 
 	return true;
 }
@@ -694,9 +845,41 @@ static bool sae_assoc_timeout(struct auth_proto *ap)
  * sent. The new SAE Commit message shall be transmitted to the peer, Sync shall
  * be zeroed, and the t0 (retransmission) timer shall be set.
  */
-static void sae_process_anti_clogging(struct sae_sm *sm, const uint8_t *ptr,
+static int sae_process_anti_clogging(struct sae_sm *sm, const uint8_t *ptr,
 					size_t len)
 {
+	/*
+	 * 802.11 doesn't talk about validating the group of the Anti-Clogging
+	 * Request message.  We assume here that the group is something that
+	 * we would have potentially sent
+	 */
+	if (len < 2)
+		return -EBADMSG;
+
+	if (sae_valid_group(sm, l_get_le16(ptr)) < 0)
+		return -EBADMSG;
+
+	len -= 2;
+	ptr += 2;
+
+	/*
+	 * 802.11-2020, Table 9-41:
+	 * When the hash-to-element method is used to derive the PWE, the
+	 * Anti-Clogging Token Container element is present if the
+	 * Status Code field is ANTI_CLOGGING_TOKEN_REQUIRED
+	 */
+	if (sm->sae_type != CRYPTO_SAE_LOOPING) {
+		if (len < 3)
+			return -EBADMSG;
+
+		if (ptr[0] != IE_TYPE_EXTENSION || ptr[2] != 93 ||
+				ptr[1] < 2 || len < ptr[1] + 2u)
+			return -EBADMSG;
+
+		len = ptr[1] - 1;
+		ptr += 3;
+	}
+
 	/*
 	 * IEEE 802.11-2016 - Section 12.4.6 Anti-clogging tokens
 	 *
@@ -707,16 +890,19 @@ static void sae_process_anti_clogging(struct sae_sm *sm, const uint8_t *ptr,
 	 * going to be 2 bytes less than the passed in length. This is why we
 	 * are checking 3 > len > 258.
 	 */
-	if (len < 3 || len > 258) {
+	if (len < 1 || len > 256) {
 		l_error("anti-clogging token size invalid %zu", len);
-		return;
+		return -EBADMSG;
 	}
 
-	sm->token = l_memdup(ptr + 2, len - 2);
-	sm->token_len = len - 2;
+	l_free(sm->token);
+	sm->token = l_memdup(ptr, len);
+	sm->token_len = len;
 	sm->sync = 0;
 
 	sae_send_commit(sm, true);
+
+	return -EAGAIN;
 }
 
 /*
@@ -742,34 +928,13 @@ static int sae_verify_nothing(struct sae_sm *sm, uint16_t transaction,
 		return -EBADMSG;
 
 	/* reject with unsupported group */
-	if (l_get_le16(frame) != sm->group) {
-		sae_reject_authentication(sm,
+	if (l_get_le16(frame) != sm->group)
+		return sae_reject(sm, SAE_STATE_COMMITTED,
 				MMPDU_STATUS_CODE_UNSUPP_FINITE_CYCLIC_GROUP);
-		return -EBADMSG;
-	}
 
 	return 0;
 }
 
-static void sae_reset_state(struct sae_sm *sm)
-{
-	l_free(sm->token);
-	sm->token = NULL;
-
-	l_ecc_scalar_free(sm->scalar);
-	sm->scalar = NULL;
-	l_ecc_scalar_free(sm->p_scalar);
-	sm->p_scalar = NULL;
-	l_ecc_scalar_free(sm->rand);
-	sm->rand = NULL;
-	l_ecc_point_free(sm->element);
-	sm->element = NULL;
-	l_ecc_point_free(sm->p_element);
-	sm->p_element = NULL;
-	l_ecc_point_free(sm->pwe);
-	sm->pwe = NULL;
-}
-
 /*
  * 802.11-2016 - 12.4.8.6.4 Protocol instance behavior - Committed state
  */
@@ -777,6 +942,9 @@ static int sae_verify_committed(struct sae_sm *sm, uint16_t transaction,
 					uint16_t status, const uint8_t *frame,
 					size_t len)
 {
+	unsigned int skip;
+	struct ie_tlv_iter iter;
+
 	/*
 	 * Upon receipt of a Con event...
 	 * Then the protocol instance checks the value of Sync. If it
@@ -788,20 +956,18 @@ static int sae_verify_committed(struct sae_sm *sm, uint16_t transaction,
 	 */
 	if (transaction == SAE_STATE_CONFIRMED) {
 		if (sm->sync > SAE_SYNC_MAX)
-			return -EBADMSG;
+			return -ETIMEDOUT;
 
 		sm->sync++;
-
 		sae_send_commit(sm, true);
 
 		return -EAGAIN;
 	}
 
-	switch (status) {
-	case MMPDU_STATUS_CODE_ANTI_CLOGGING_TOKEN_REQ:
-		sae_process_anti_clogging(sm, frame, len);
-		return -EAGAIN;
-	case MMPDU_STATUS_CODE_UNSUPP_FINITE_CYCLIC_GROUP:
+	if (status == MMPDU_STATUS_CODE_ANTI_CLOGGING_TOKEN_REQ)
+		return sae_process_anti_clogging(sm, frame, len);
+
+	if (status == MMPDU_STATUS_CODE_UNSUPP_FINITE_CYCLIC_GROUP) {
 		/*
 		 * TODO: hostapd in its current state does not include the
 		 * group number as it should. This is a violation of the spec,
@@ -822,19 +988,9 @@ static int sae_verify_committed(struct sae_sm *sm, uint16_t transaction,
 		if (len == 0)
 			l_warn("AP did not include group number in response!");
 		else if (len >= 2 && (l_get_le16(frame) != sm->group))
-			return -EBADMSG;
+			return -ENOMSG;
 
-		sm->group_retry++;
-
-		if (sm->ecc_groups[sm->group_retry] == 0) {
-			/*
-			 * "If there are no other groups to choose, the protocol
-			 * instance shall send a Del event to the parent process
-			 * and transitions back to Nothing state"
-			 */
-			sm->state = SAE_STATE_NOTHING;
-			goto reject_unsupp_group;
-		}
+		sae_rejected_groups_append(sm, L_CPU_TO_LE16(sm->group));
 
 		/*
 		 * "If the rejected group matches the last offered group, the
@@ -844,97 +1000,88 @@ static int sae_verify_committed(struct sae_sm *sm, uint16_t transaction,
 		 * zeros Sync, sets the t0 (retransmission) timer, and remains
 		 * in Committed state"
 		 */
-
-		sae_reset_state(sm);
-
-		sm->group = sm->ecc_groups[sm->group_retry];
-		sm->curve = l_ecc_curve_get_ike_group(sm->group);
-
-		sm->sync = 0;
-
-		sae_send_commit(sm, false);
-
-		return -EAGAIN;
-	case 0:
-		if (len < 2)
-			return -EBADMSG;
-
-		if (l_get_le16(frame) == sm->group)
-			return 0;
-
-		if (!l_ecc_curve_get_ike_group(l_get_le16(frame))) {
-			if (sm->sync > SAE_SYNC_MAX)
-				return -EBADMSG;
-
-			sm->sync++;
-
+		if (sae_choose_next_group(sm) < 0) {
+			/*
+			 * "If there are no other groups to choose, the protocol
+			 * instance shall send a Del event to the parent process
+			 * and transitions back to Nothing state"
+			 */
+			sm->state = SAE_STATE_NOTHING;
 			goto reject_unsupp_group;
 		}
 
-		/*
-		 * If we get here we know that the groups do not match, but the
-		 * group provided in the frame is supported. From section
-		 * 12.4.8.6.4 we see:
-		 *
-		 * "If the group is supported but does not match that used when
-		 * the protocol instance constructed its SAE Commit message,
-		 * DiffGrp shall be set and the local identity and peer identity
-		 * shall be checked"
-		 */
-
-		if (memcmp(sm->handshake->spa, sm->handshake->aa, 6) > 0) {
-			/*
-			 * "The mesh STA, with the numerically greater of the two
-			 * MAC addresses, drops the received SAE Commit message,
-			 * retransmits its last SAE Commit message, and shall
-			 * set the t0 (retransmission) timer and remain in
-			 * Committed state"
-			 */
-			sae_send_commit(sm, true);
-
-			return 0;
-		}
-
-		/*
-		 * "The mesh STA, with the numerically lesser of the two
-		 * MAC addresses, zeros Sync, shall increment Sc, choose
-		 * the group from the received SAE Commit message,
-		 * generate new PWE and new secret values according to
-		 * 12.4.5.2, process the received SAE Commit message
-		 * according to 12.4.5.4, generate a new SAE Commit
-		 * message and SAE Confirm message, and shall transmit
-		 * the new Commit and Confirm to the peer. It shall then
-		 * transition to Confirmed state."
-		 */
 		sm->sync = 0;
-		sm->sc++;
-		sm->group = l_get_le16(frame);
-		sm->curve = l_ecc_curve_get_ike_group(sm->group);
-
 		sae_send_commit(sm, false);
 
-		sm->state = SAE_STATE_CONFIRMED;
-
-		/*
-		 * The processing and sending of the confirm message
-		 * will happen after we return. Since we have set the
-		 * state to CONFIRMED, our confirm handler will get
-		 * called.
-		 */
-
-		return 0;
-	default:
-		/*
-		 * If the Status is some other nonzero value, the frame shall
-		 * be silently discarded...
-		 */
-		return 0;
+		return -EAGAIN;
 	}
 
+	/*
+	 * If the Status is some other nonzero value, the frame shall be
+	 * silently discarded and the t0 (retransmission) timer shall be set.
+	 */
+	if (status != 0 && status != MMPDU_STATUS_CODE_SAE_HASH_TO_ELEMENT)
+		return -ENOMSG;
+
+	if (status != sae_status_code(sm))
+		return -EBADMSG;
+
+	if (len < 2)
+		return -EBADMSG;
+
+	if (l_get_le16(frame) != sm->group) {
+		l_error("SAE: Peer tried to change group -- Reject");
+		goto reject_unsupp_group;
+	}
+
+	len -= 2;
+	frame += 2;
+
+	skip = l_ecc_curve_get_scalar_bytes(sm->curve) * 3;
+	if (len < skip)
+		return -EBADMSG;
+
+	/* If H2E isn't being used, there should be no IEs in use */
+	if (status == 0)
+		return 0;
+
+	len -= skip;
+	frame += skip;
+
+	ie_tlv_iter_init(&iter, frame, len);
+
+	while (ie_tlv_iter_next(&iter)) {
+		switch (ie_tlv_iter_get_tag(&iter)) {
+		/*
+		 * If the peer's SAE Commit message contains a Rejected Groups
+		 * element, the list of rejected groups shall be checked to
+		 * ensure that all of the groups in the list are groups that
+		 * would be rejected. If any groups in the list would not be
+		 * rejected then processing of the SAE Commit message
+		 * terminates and the STA shall reject the peer's
+		 * authentication.
+		 *
+		 * NOTE: We currently only support the Initiator role, and so
+		 * do not reject any groups.  We should never receive this
+		 * element
+		 */
+		case IE_TYPE_REJECTED_GROUPS:
+			l_error("SAE: Unexpected Rejected Groups IE");
+			return sae_reject(sm, SAE_STATE_COMMITTED,
+					MMPDU_STATUS_CODE_UNSPECIFIED);
+		/* We don't request tokens, so we shouldn't get any */
+		case IE_TYPE_ANTI_CLOGGING_TOKEN_CONTAINER:
+			l_error("SAE: Unexpected Anti-Clogging Container IE");
+			return sae_reject(sm, SAE_STATE_COMMITTED,
+					MMPDU_STATUS_CODE_UNSPECIFIED);
+		}
+	}
+
+	return 0;
+
 reject_unsupp_group:
-	sae_reject_authentication(sm,
+	return sae_reject(sm, SAE_STATE_COMMITTED,
 			MMPDU_STATUS_CODE_UNSUPP_FINITE_CYCLIC_GROUP);
-	return MMPDU_STATUS_CODE_UNSUPP_FINITE_CYCLIC_GROUP;
 }
 
 /*
@@ -944,14 +1091,35 @@ static int sae_verify_confirmed(struct sae_sm *sm, uint16_t trans,
 					uint16_t status, const uint8_t *frame,
 					size_t len)
 {
-	if (trans == SAE_STATE_CONFIRMED)
+	if (trans == SAE_STATE_CONFIRMED) {
+		enum l_checksum_type hash =
+			crypto_sae_hash_from_ecc_prime_len(sm->sae_type,
+				l_ecc_curve_get_scalar_bytes(sm->curve));
+		size_t hash_len = l_checksum_digest_length(hash);
+
+		/* Most likely the password is wrong */
+		if (status == MMPDU_STATUS_CODE_UNSPECIFIED && len == 0)
+			return -ENOKEY;
+
+		if (status != MMPDU_STATUS_CODE_SUCCESS)
+			return -EPROTO;
+
+		if (len < hash_len + 2) {
+			l_error("SAE: Confirm packet too short");
+			return -EBADMSG;
+		}
+
 		return 0;
+	}
 
 	/*
-	 * If the Status is nonzero, the frame shall be silently discarded...
+	 * Upon receipt of a Com event, the t0 (retransmission) timer shall be
+	 * canceled. If the Status is nonzero, the frame shall be silently
+	 * discarded, the t0 (retransmission) timer set, and the protocol
+	 * instance shall remain in the Confirmed state.
 	 */
 	if (status != 0)
-		return 0;
+		return -ENOMSG;
 
 	/*
 	 * If Sync is greater than dot11RSNASAESync, the protocol instance
@@ -959,14 +1127,14 @@ static int sae_verify_confirmed(struct sae_sm *sm, uint16_t trans,
 	 * Nothing state.
 	 */
 	if (sm->sync > SAE_SYNC_MAX)
-		return -EBADMSG;
+		return -ETIMEDOUT;
 
 	if (len < 2)
 		return -EBADMSG;
 
 	/* frame shall be silently discarded */
 	if (l_get_le16(frame) != sm->group)
-		return 0;
+		return -EBADMSG;
 
 	/*
 	 * the protocol instance shall increment Sync, increment Sc, and
@@ -976,9 +1144,11 @@ static int sae_verify_confirmed(struct sae_sm *sm, uint16_t trans,
 	sm->sc++;
 
 	sae_send_commit(sm, true);
-	sae_send_confirm(sm);
 
-	return 0;
+	if (!sae_send_confirm(sm))
+		return -EPROTO;
+
+	return -EAGAIN;
 }
 
 /*
@@ -997,7 +1167,7 @@ static int sae_verify_accepted(struct sae_sm *sm, uint16_t trans,
 	}
 
 	if (sm->sync > SAE_SYNC_MAX)
-		return -EBADMSG;
+		return -ETIMEDOUT;
 
 	if (len < 2)
 		return -EBADMSG;
@@ -1028,9 +1198,10 @@ static int sae_verify_accepted(struct sae_sm *sm, uint16_t trans,
 	sm->sync++;
 	sm->sc = 0xffff;
 
-	sae_send_confirm(sm);
+	if (!sae_send_confirm(sm))
+		return -EPROTO;
 
-	return 0;
+	return -EAGAIN;
 }
 
 static int sae_verify_packet(struct sae_sm *sm, uint16_t trans,
@@ -1052,24 +1223,17 @@ static int sae_verify_packet(struct sae_sm *sm, uint16_t trans,
 	}
 
 	/* should never get here */
-	return -1;
+	return -EPROTO;
 }
 
 static int sae_rx_authenticate(struct auth_proto *ap,
 				const uint8_t *frame, size_t len)
 {
 	struct sae_sm *sm = l_container_of(ap, struct sae_sm, ap);
-	const struct mmpdu_header *hdr = mpdu_validate(frame, len);
-	const struct mmpdu_authentication *auth;
+	const struct mmpdu_header *hdr = (const struct mmpdu_header *) frame;
+	const struct mmpdu_authentication *auth = mmpdu_body(hdr);
 	int ret;
 
-	if (!hdr) {
-		l_debug("Auth frame header did not validate");
-		goto reject;
-	}
-
-	auth = mmpdu_body(hdr);
-
 	len -= mmpdu_header_len(hdr);
 
 	ret = sae_verify_packet(sm, L_LE16_TO_CPU(auth->transaction_sequence),
@@ -1090,28 +1254,18 @@ static int sae_rx_authenticate(struct auth_proto *ap,
 				L_LE16_TO_CPU(auth->transaction_sequence));
 	}
 
-reject:
-	sae_reject_authentication(sm, MMPDU_REASON_CODE_UNSPECIFIED);
-
-	return -EBADMSG;
+	/* should never get here */
+	return -EPROTO;
 }
 
 static int sae_rx_associate(struct auth_proto *ap, const uint8_t *frame,
 				size_t len)
 {
-	const struct mmpdu_header *mpdu = NULL;
-	const struct mmpdu_association_response *body;
-
-	mpdu = mpdu_validate(frame, len);
-	if (!mpdu) {
-		l_error("could not process frame");
-		return -EBADMSG;
-	}
-
-	body = mmpdu_body(mpdu);
+	const struct mmpdu_header *mpdu = (const struct mmpdu_header *)frame;
+	const struct mmpdu_association_response *body = mmpdu_body(mpdu);
 
 	if (body->status_code != 0)
-		return L_LE16_TO_CPU(body->status_code);
+		return -EPROTO;
 
 	return 0;
 }
@@ -1125,15 +1279,37 @@ static bool sae_start(struct auth_proto *ap)
 	else
 		memcpy(sm->peer, sm->handshake->aa, 6);
 
+	if (sm->sae_type == CRYPTO_SAE_LOOPING && !sm->handshake->passphrase) {
+		l_error("SAE: No passphrase set");
+		return false;
+	}
+
+	if (sae_choose_next_group(sm) < 0)
+		return false;
+
+	sm->state = SAE_STATE_COMMITTED;
 	return sae_send_commit(sm, false);
 }
 
+bool sae_sm_is_h2e(struct auth_proto *ap)
+{
+	struct sae_sm *sm = l_container_of(ap, struct sae_sm, ap);
+
+	return sm->sae_type != CRYPTO_SAE_LOOPING;
+}
+
 static void sae_free(struct auth_proto *ap)
 {
 	struct sae_sm *sm = l_container_of(ap, struct sae_sm, ap);
 
 	sae_reset_state(sm);
 
+	l_free(sm->token);
+	sm->token = NULL;
+
+	if (sm->rejected_groups)
+		free(sm->rejected_groups);
+
 	/* zero out whole structure, including keys */
 	explicit_bzero(sm, sizeof(struct sae_sm));
 
@@ -1146,20 +1322,17 @@ struct auth_proto *sae_sm_new(struct handshake_state *hs,
 				void *user_data)
 {
 	struct sae_sm *sm;
+	const void *rsnxe;
 
 	sm = l_new(struct sae_sm, 1);
 
-	if (!sm)
-		return NULL;
+	sm->group_retry = -1;
 
 	sm->tx_auth = tx_auth;
 	sm->tx_assoc = tx_assoc;
 	sm->user_data = user_data;
 	sm->handshake = hs;
 	sm->state = SAE_STATE_NOTHING;
-	sm->ecc_groups = l_ecc_curve_get_supported_ike_groups();
-	sm->group = sm->ecc_groups[sm->group_retry];
-	sm->curve = l_ecc_curve_get_ike_group(sm->group);
 
 	sm->ap.start = sae_start;
 	sm->ap.free = sae_free;
@@ -1167,5 +1340,16 @@ struct auth_proto *sae_sm_new(struct handshake_state *hs,
 	sm->ap.rx_associate = sae_rx_associate;
 	sm->ap.assoc_timeout = sae_assoc_timeout;
 
+	rsnxe = hs->authenticator ? hs->supplicant_rsnxe :
+						hs->authenticator_rsnxe;
+
+	if (ie_rsnxe_capable(rsnxe, IE_RSNX_SAE_H2E) && hs->ecc_sae_pts) {
+		l_debug("Using SAE H2E");
+		sm->sae_type = CRYPTO_SAE_HASH_TO_ELEMENT;
+	} else {
+		l_debug("Using SAE Hunting and Pecking");
+		sm->sae_type = CRYPTO_SAE_LOOPING;
+	}
+
 	return &sm->ap;
 }
